# -*- coding: utf-8 -*-

import constants
import os
import utils


class ServiceState(object):
    def __init__(self, pid_file_path=constants.DEFAULT_PID_FILE_LOCATION):
        self.job_worker_pids = []
        self.retry_worker_pids = []
        self.dlq_worker_pids = []
        self.pid_file = pid_file_path
        if os.path.exists(self.pid_file):
            tmp_obj = utils.load_obj(self.pid_file)
            self.__dict__ = tmp_obj.__dict__
        else:
            self.persist_state()

    def persist_state(self):
        """
        Persist the library pid state to the pid file
        :return: 
        """
        utils.save_obj(self, self.pid_file)

    def get_pid_list(self, type):
        """
        Get the pid list for the give type of worker
        :param type: type of worker
        :return: 
        """
        if type == constants.WORK_QUEUE:
            return self.job_worker_pids
        elif type == constants.RETRY_QUEUE:
            return self.retry_worker_pids
        elif type == constants.DEAD_LETTER_QUEUE:
            return self.dlq_worker_pids
        else:
            raise KeyError("Invalid key : " + str(type))

    def add_worker_pid(self, type, pid):
        """
        add a pid for the worker
        :param type: type of worker
        :param pid: pid to be added
        :return: 
        """
        if type == constants.WORK_QUEUE:
            self.job_worker_pids.append(pid)
        elif type == constants.RETRY_QUEUE:
            self.retry_worker_pids.append(pid)
        elif type == constants.DEAD_LETTER_QUEUE:
            self.dlq_worker_pids.append(pid)
        else:
            raise KeyError("Invalid key : " + str(type))
        self.persist_state()

    def remove_worker_pid(self, type, pid):
        """
        remove a worker pid from the list
        :param type: type of worker
        :param pid: pid to be removed
        :return: 
        """
        if type == constants.WORK_QUEUE:
            self.job_worker_pids.remove(pid)
        elif type == constants.RETRY_QUEUE:
            self.retry_worker_pids.remove(pid)
        elif type == constants.DEAD_LETTER_QUEUE:
            self.dlq_worker_pids.remove(pid)
        else:
            raise KeyError("Invalid key : " + str(type))
        self.persist_state()



# -*- coding: utf-8 -*-

import importlib
import inspect
import logging
import pickle
import traceback

import constants
import exception
import requests
from response import EasyResponse
from utils import as_text, is_string_type

logging.basicConfig()


class EasyApi(object):
    def __init__(self):
        self.type = constants.API_REMOTE
        self._instance = None
        self._func_name = None
        self.api_request_headers = {}
        self.remote_call_type = "post"
        self.content_type = "application/json"
        self._data = None

    @classmethod
    def create(cls, type, api, api_request_headers={}, remote_call_type=None, content_type=None):
        if type not in constants.job_call_types:
            raise TypeError("Invalid remote_call_type: {}".format(type))
        easy_api = cls()
        easy_api.type = type
        easy_api._instance = None
        easy_api._func_name = None
        easy_api.api_request_headers = api_request_headers
        if remote_call_type:
            easy_api.remote_call_type = remote_call_type
        if content_type:
            easy_api.content_type = content_type

        if easy_api.type == constants.API_LOCAL:
            if inspect.ismethod(api):
                easy_api._instance = api.__self__
                easy_api._func_name = api.__name__
            elif inspect.isfunction(api) or inspect.isbuiltin(api):
                easy_api._func_name = '{0}.{1}'.format(api.__module__, api.__name__)
            elif is_string_type(api):
                easy_api._func_name = as_text(api)
            elif not inspect.isclass(api) and hasattr(api, '__call__'):  # a callable class instance
                easy_api._instance = api
                easy_api._func_name = '__call__'
            else:
                raise TypeError('Expected a callable or a string for local job type, but got: {0}'.format(api))
        else:
            easy_api._func_name = api
        job_tuple = easy_api._func_name, easy_api._instance
        easy_api._data = pickle.dumps(job_tuple)
        return easy_api

    def load_data(self):
        return pickle.loads(self._data)

    @classmethod
    def create_from_dict(cls, dict_data):
        if not dict_data:
            return None

        easy_api = cls()
        if "type" in dict_data:
            easy_api.type = dict_data["type"]
        if "data" in dict_data:
            easy_api._data = dict_data["data"]
            easy_api._func_name, easy_api._instance = easy_api.load_data()
        if "api_request_headers" in dict_data:
            easy_api.api_request_headers = dict_data["api_request_headers"]
        if "remote_call_type" in dict_data:
            easy_api.remote_call_type = dict_data["remote_call_type"]
        if "content_type" in dict_data:
            easy_api.content_type = dict_data["content_type"]

        return easy_api

    def to_dict(self):
        dict_data = {
            "type": self.type,
            "instance": pickle.dumps(self._instance, protocol=0),
            "data": self._data,
            "api_request_headers": self.api_request_headers,
            "remote_call_type": self.remote_call_type,
            "content_type": self.content_type
        }
        return dict_data

    @property
    def api(self):
        return self._func_name

    @property
    def func(self):
        func_name = self._func_name
        if func_name is None:
            return None

        if self._instance:
            return getattr(self._instance, func_name)

        module_name, attribute = func_name.rsplit('.', 1)
        module = importlib.import_module(module_name)
        return getattr(module, attribute)

    def _call_remote(self, api, call_type, data, async_timeout):
        request_call_type = constants.remote_call_type.get(call_type)
        request_headers = self.api_request_headers
        request_headers['Content-type'] = self.content_type
        if call_type == 'post':
            response = request_call_type(api, data=data, timeout=async_timeout,
                                         headers=request_headers)
        else:
            response = request_call_type(api, timeout=async_timeout,
                                         headers=request_headers)
        return response

    def _call_local(self, func_call, data):
        return func_call(data)

    def execute(self, data, async_timeout=constants.DEFAULT_ASYNC_TIMEOUT):
        """
        execute the api call
        :param data: data to be sent while executing
        :param async_timeout: time out for the api (honoured only in case of remote)
        :return: 
        """
        logger = logging.getLogger(self.__class__.__name__)
        ret_val = EasyResponse(200, "")
        try:
            if self.type == constants.API_REMOTE:
                try:
                    call_type = self.remote_call_type
                    response = self._call_remote(self.api, call_type, data, async_timeout)

                except requests.exceptions.Timeout as e:
                    raise exception.ApiTimeoutException(self.api, async_timeout)

                ret_val.status_code = response.status_code
                ret_val.message = response.text

            elif self.type == constants.API_LOCAL:
                response = self._call_local(self.func, data)
                if response and not is_string_type(response):
                    if hasattr(response, 'status_code'):
                        ret_val.status_code = response.status_code
                    if hasattr(response, 'message'):
                        ret_val.message = response.message
                    if hasattr(response, 'data'):
                        ret_val.data = response.data
                    else:
                        ret_val.data = response
                logger.info("Calling api {api} with {data}".format(api=self.api, data=data))
        except Exception as e:
            traceback.print_exc()
            error_message = "Error running API '{api}': {err}".format(api=self.api, err=e.__class__.__name__)
            logger.error(error_message)
            ret_val.status_code = 400
            ret_val.message = error_message
            ret_val.data = e.__dict__
        return ret_val



# -*- coding: utf-8 -*-

import logging
import traceback

import constants
import yaml
from easyjoblite.utils import update_import_paths


class Configuration(object):
    """
    used to configure the service
    """

    def __init__(self, **kwargs):
        """
        :param app_id: the application specific unique name.
        :param rabbitmq_url: where it the rabbitmq server running?
        :param max_retries: max retries before a message moves to DLQ
        :param async_timeout: wait these many seconds before we timeout async reqs
        :param eqc_sleep_duration: how long (in mins) should error-queue-consumer sleep between bursts
        :param max_worker_count: max ammount of job workers
        :param default_worker_count: default about of job worker to be spawned
        :param default_retry_consumer_count: default amount of retry consumer to be spawned
        :param default_dl_consumer_count: default amount of dead letter consumer to be spawned
        :param import_paths: the paths for looking to import modules for local run
        :param pid_file_path: the path to store the pid files of all the workers
        :param workers_log_file_path: the path to log from workers
        :param dead_message_log_file: the path to dump the dead letter messages without a event handler
        :param config_file: the path to store configuration
        """

        # first initialize everything with defaults
        self.load_with_defaults()

        # then check if config file is there if yes then load from it
        if 'config_file' in kwargs:
            self.config_file = kwargs['config_file']

        self.load_from_file(self.config_file)

        # now load any run time config change done
        self.set_config(**kwargs)

    def load_with_defaults(self):
        self.app_id = constants.DEFAULT_APP_ID
        self.rabbitmq_url = constants.DEFAULT_RMQ_URL
        self.max_retries = constants.DEFAULT_MAX_JOB_RETRIES
        self.async_timeout = constants.DEFAULT_ASYNC_TIMEOUT
        self.eqc_sleep_duration = constants.DEFAULT_ERROR_Q_CON_SLEEP_DURATION
        self.max_worker_count = constants.DEFAULT_MAX_WORKER_COUNT
        self.default_worker_count = constants.DEFAULT_WORKER_COUNT
        self.default_retry_consumer_count = constants.DEFAULT_RETRY_CONSUMER_COUNT
        self.default_dl_consumer_count = constants.DEFAULT_DLQ_CONSUMER_COUNT
        self.import_paths = constants.DEFAULT_IMPORT_PATHS
        self.pid_file_path = constants.DEFAULT_PID_FILE_LOCATION
        self.workers_log_file_path = constants.DEFAULT_LOG_FILE_PATH
        self.dead_message_log_file = constants.DEFAULT_DL_LOG_FILE
        self.config_file = constants.DEFAULT_CONFIG_FILE
        update_import_paths(self.import_paths)

    def set_config(self, **kwargs):
        """
        set config values
        :param kwargs: contains the dict with all key values
        :return: 
        """
        if 'app_id' in kwargs:
            self.app_id = kwargs['app_id']

        if 'rabbitmq_url' in kwargs:
            self.rabbitmq_url = kwargs['rabbitmq_url']

        if 'async_timeout' in kwargs:
            self.async_timeout = kwargs['async_timeout']

        if 'max_retries' in kwargs:
            self.max_retries = kwargs['max_retries']

        if 'eqc_sleep_duration' in kwargs:
            self.eqc_sleep_duration = kwargs['eqc_sleep_duration']

        if 'import_paths' in kwargs:
            self.import_paths = kwargs['import_paths']
            update_import_paths(self.import_paths)

        if 'max_worker_count' in kwargs:
            self.max_worker_count = kwargs['max_worker_count']

        if 'default_worker_count' in kwargs:
            self.default_worker_count = kwargs['default_worker_count']

        if 'default_retry_consumer_count' in kwargs:
            self.default_retry_consumer_count = kwargs['default_retry_consumer_count']

        if 'default_dl_consumer_count' in kwargs:
            self.default_dl_consumer_count = kwargs['default_dl_consumer_count']

        if 'workers_log_file_path' in kwargs:
            self.workers_log_file_path = kwargs['workers_log_file_path']

        if 'dead_message_log_file' in kwargs:
            self.dead_message_log_file = kwargs['dead_message_log_file']

        if 'config_file' in kwargs:
            self.config_file = kwargs['config_file']

    def load_from_file(self, file_path):
        logger = logging.getLogger(self.__class__.__name__)
        try:
            with open(file_path, 'r') as stream:
                loaded_config = yaml.load(stream)
                self.set_config(**loaded_config)
        except Exception as e:
            traceback.print_exc()
            logger.error("Unable to load config file: {0} with exception {1}".format(str(file_path), e.message))
            self.load_with_defaults()
            self.dump_to_file(file_path)

    def dump_to_file(self, file_path):
        logger = logging.getLogger(self.__class__.__name__)
        conf_dict = {
            "app_id": self.app_id,
            "rabbitmq_url": self.rabbitmq_url,
            "max_retries": self.max_retries,
            "async_timeout": self.async_timeout,
            "eqc_sleep_duration": self.eqc_sleep_duration,
            "max_worker_count": self.max_worker_count,
            "default_worker_count": self.default_worker_count,
            "default_retry_consumer_count": self.default_retry_consumer_count,
            "default_dl_consumer_count": self.default_dl_consumer_count,
            "import_paths": self.import_paths,
            "pid_file_path": self.pid_file_path,
            "workers_log_file_path": self.workers_log_file_path,
            "dead_message_log_file": self.dead_message_log_file
        }

        try:
            config_file = open(file_path, "w")
            yaml.safe_dump(conf_dict, config_file, default_flow_style=False)
            config_file.close()
        except Exception as e:
            traceback.print_exc()
            logger.error("Unable to dump config file: {0} with exception {1}".format(str(file_path), e.message))

    def get_mq_config(self, conf_type):
        """
        return the rabbitmq config type
        :param conf_type: type like EXCHANGE, WORK_QUEUE, RETRY_QUEUE etc
        :return: 
        """
        return self.app_id + constants.rabbit_mq_prefix[conf_type]




# -*- coding: utf-8 -*-

import logging
import signal
import socket
import traceback

from easyjoblite.utils import enqueue
from kombu import Connection
from kombu import Consumer
from kombu import Producer


class BaseRMQConsumer(object):
    """
    base class for all rabbit-mq workers that consume from any of the job queues
    """

    def __init__(self, orchestrator):
        """
        :param configuration: all necessary configuration needed for the worker
        """
        logger = logging.getLogger(self.__class__.__name__)

        try:
            self._orchestrator = orchestrator
            # todo: do we need to make confirm_publish configurable?
            self._conn = Connection(self.get_config().rabbitmq_url,
                                    transport_options={'confirm_publish': True})

            # setup producer to push to error and dlqs
            self._producer = Producer(channel=self._conn.channel(),
                                      exchange=self._orchestrator.get_exchange())
            signal.signal(signal.SIGTERM, self.signal_term_handler)
            self._should_block = True

        except Exception as e:
            traceback.print_exc()
            logger.error("Error connecting to rabbitmq({u}): {err}".format(u=self.get_config().rabbitmq_url,
                                                                           err=e.message))
            raise

    def consume(self, from_queue, blocking=True):
        """
        consumes from from_queue and feeds each of the message into
        process_message callback. this is a blocking call

        todo: this can later be ended using a poison-pill

        :param from_queue: kombu.Queue to consume from
        :param blocking: should it block waiting for new messages?
        :return: nothing, might raise exception/s
        """

        logger = logging.getLogger(self.__class__.__name__)
        channel = self._conn.channel()

        # prep a consumer for the from_queue only
        queue_consumer = Consumer(channel=channel,
                                  queues=[from_queue],
                                  callbacks=[self.process_message])
        queue_consumer.consume()

        try:
            # drain the events into the consumer
            logger.info("starting to consume messages from {exchg}:{q}".format(exchg=from_queue.exchange.name,
                                                                               q=from_queue.name))
            while self.should_run_loop():
                if blocking:
                    self._conn.drain_events()

                else:
                    # at max we wait for 5 more seconds after exhausting all work-items from
                    # the queue.
                    try:
                        self._conn.drain_events(timeout=1)

                    except socket.timeout:
                        logger.debug("No more work-items in {q}".format(q=from_queue.name))
                        break

                    except socket.error as e:
                        # we don't care about EAGAIN, since we had intentionally started a non-blocking conn
                        if e.errno == 35:
                            msg = "{q} is empty".format(q=from_queue.name)
                            logger.debug(msg)

                        else:
                            traceback.print_exc()
                            logger.error(e)

                        break

        except Exception as e:
            traceback.print_exc()
            logger.error("Something broke while listening for new messages: {err}".format(err=e.message))
            logger.info("Quitting")

        # disconnect from the queue
        queue_consumer.cancel()

    def process_message(self, body, message):
        """
        gets called back once a message arrives in the work queue

        :param body: message payload
        :param message: queued message with headers and other metadata
        """
        raise NotImplementedError("'{n}' needs to implement process_message(...)".format(n=self.__class__.__name__))

    def produce_to_queue(self, type, body, job):
        enqueue(self._producer, type, job, body)

    def get_config(self):
        return self._orchestrator.get_config()

    def should_run_loop(self):
        return self._should_block

    def signal_term_handler(self, signum, frame):
        logger = logging.getLogger(self.__class__.__name__)
        logger.debug("SIGTERM found so stopping worker")
        self._should_block = False
